# -*- coding: utf-8 -*-

###########################################################################
# Glest Model / Texture / UV / Animation Importer and Exporter
# for the Game Glest that u can find http://www.glest.org
# copyright 2005 By Andreas Becker (seltsamuel@yahoo.de)
#
# 2011/05/25: v0.1 alpha1
# modified by William Zheng for Blender 2.57(loveheaven_zhengwei@hotmail.com)
#
# corrected by MrPostiga for Blender 2.58
#
# extended by Yggdrasil
#
# Started Date: 07 June 2005  Put Public 20 June 2005   
#  Distributed under the GNU PUBLIC LICENSE
#"""
#Here an explanation of the V4 Format found at www.glest.org
#================================
#1. DATA TYPES
#================================
#G3D files use the following data types:
#uint8: 8 bit unsigned integer
#uint16: 16 bit unsigned integer
#uint32: 32 bit unsigned integer
#float32: 32 bit floating point
#================================
#2. OVERALL STRUCTURE
#================================
#- File header
#- Model header
#- Mesh header
#- Texture names
#- Mesh data
#================================
#2. FILE HEADER
#================================
#Code:
#struct FileHeader{
#   uint8 id[3];
#   uint8 version;
#};
#This header is shared among all the versions of G3D, it identifies this file as a G3D model and provides information of the version.
#id: must be "G3D"
#version: must be 4, in binary (not '4')
#================================
#3. MODEL HEADER
#================================
#Code:
#struct ModelHeader{
#   uint16 meshCount;
#   uint8 type;
#};		   
#meshCount: number of meshes in this model
#type: must be 0
#================================
#4. MESH HEADER
#================================
#There is a mesh header for each mesh, there must be "meshCount" headers in a file but they are not consecutive, texture names and mesh data are stored in between.
#Code:
#struct MeshHeader{
#   uint8 name[64];
#   uint32 frameCount;
#   uint32 vertexCount;
#   uint32 indexCount;
#   float32 diffuseColor[3];
#   float32 specularColor[3];
#   float32 specularPower;
#   float32 opacity;
#   uint32 properties;
#   uint32 textures;
#};
#name: name of the mesh
#frameCount: number of keyframes in this mesh
#vertexCount: number of vertices in each frame
#indexCount: number of indices in this mesh (the number of triangles is indexCount/3)
#diffuseColor: RGB diffuse color
#specularColor: RGB specular color (currently unused)
#specularPower: specular power (currently unused)
##properties: property flags
#Code:
#enum MeshPropertyFlag{
#		mpfCustomColor= 1,
#		mpfTwoSided= 2,
#		mpfNoSelect= 4
#};
#mpfTwoSided: meshes in this mesh are rendered by both sides, if this flag is not present only "counter clockwise" faces are rendered
#mpfCustomColor: alpha in this model is replaced by a custom color, usually the player color
#textures: texture flags
#Code:
#enum MeshTexture{
#		diffuse = 1,
#		specular = 2,
#		normal = 4
#};
#================================
#4. TEXTURE NAMES
#================================
#A list of uint8[64] texture name values. One for each texture in the mesh. If there are no textures in the mesh no texture names are present.
#================================
#5. MESH DATA
#================================
#After each mesh header and texture names the mesh data is placed.
#vertices: frameCount * vertexCount * 3, float32 values representing the x, y, z vertex coords for all frames
#normals: frameCount * vertexCount * 3, float32 values representing the x, y, z normal coords for all frames
#texture coords: vertexCount * 2, float32 values representing the s, t tex coords for all frames (only present if the mesh has 1 texture at least)
#indices: indexCount, uint32 values representing the indices
###########################################################################

bl_info = {
	"name": "G3D Mesh Import/Export",
	"description": "Import/Export .g3d file (Glest 3D)",
	"author": "various, see head of script",
	"version": (0, 11, 1),
	"blender": (2, 65, 0),
	"location": "File > Import-Export",
	"warning": "always keep .blend files",
	"wiki_url": "http://glest.wikia.com/wiki/G3D_support",
	"tracker_url": "https://forum.megaglest.org/index.php?topic=6596",
	"category": "Import-Export"}
###########################################################################
# Importing Structures needed (must later verify if i need them really all)
###########################################################################
import bpy
from bpy.props import StringProperty
from bpy_extras.image_utils import load_image
from bpy_extras.io_utils import ImportHelper, ExportHelper
import bmesh

import sys, struct, string, types
from types import *
import os
from os import path
from os.path import dirname, abspath

import subprocess
from mathutils import Matrix
from math import radians
###########################################################################
# Variables that are better Global to handle
###########################################################################
imported = []   #List of all imported Objects
toexport = []   #List of Objects to export (actually only meshes)
sceneID  = None #Points to the active Blender Scene
def unpack_list(list_of_tuples):
	l = []
	for t in list_of_tuples:
		l.extend(t)
	return l
###########################################################################
# Declaring Structures of G3D Format
###########################################################################
class G3DHeader:							#Read first 4 Bytes of file should be G3D + Versionnumber
	binary_format = "<3cB"
	def __init__(self, fileID):
		temp = fileID.read(struct.calcsize(self.binary_format))
		data = struct.unpack(self.binary_format,temp)
		self.id = str(data[0]+data[1]+data[2], "utf-8")
		self.version = data[3]

class G3DModelHeaderv3:					 #Read Modelheader in V3 there is only the number of Meshes in file
	binary_format = "<I"
	def __init__(self, fileID):
		temp = fileID.read(struct.calcsize(self.binary_format))
		data = struct.unpack(self.binary_format,temp)
		self.meshcount = data[0]

class G3DModelHeaderv4:					 #Read Modelheader: Number of Meshes and Meshtype (must be 0)
	binary_format = "<HB"
	def __init__(self, fileID):
		temp = fileID.read(struct.calcsize(self.binary_format))
		data = struct.unpack(self.binary_format,temp)
		self.meshcount = data[0]
		self.mtype = data[1]
		
class G3DMeshHeaderv3:										 #Read Meshheader
	binary_format = "<7I64c"
	def __init__(self,fileID):
		temp = fileID.read(struct.calcsize(self.binary_format))
		data = struct.unpack(self.binary_format,temp)
		self.framecount = data[0]				 #Framecount = Number of Animationsteps
		self.normalframecount= data[1]		 #Number of Normal Frames actualli equal to Framecount
		self.texturecoordframecount= data[2]#Number of Frames of Texturecoordinates seems everytime to be 1
		self.colorframecount= data[3]		  #Number of Frames of Colors seems everytime to be 1
		self.vertexcount= data[4]				  #Number of Vertices in each Frame
		self.indexcount= data[5]					   #Number of Indices in Mesh (Triangles = Indexcount/3)
		self.properties= data[6]					   #Property flags
		if self.properties & 1:					  #PropertyBit is Mesh Textured ?
			self.hastexture = False
			self.diffusetexture = None
		else:
			self.diffusetexture = "".join([str(x, "ascii") for x in data[7:-1] if x[0]< 127 ])
			self.hastexture = True
		if self.properties & 2:					  #PropertyBit is Mesh TwoSided ?
			self.istwosided = True
		else:
			self.istwosided = False
		if self.properties & 4:					  #PropertyBit is Mesh Alpha Channel custom Color in Game ?
			self.customalpha = True
		else:
			self.customalpha = False
		
class G3DMeshHeaderv4:										 #Read Meshheader
	binary_format = "<64c3I8f2I"
	texname_format = "<64c"

	def _readtexname(self,fileID):
		temp = fileID.read(struct.calcsize(self.texname_format))
		data = struct.unpack(self.texname_format,temp)
		return "".join([str(x, "ascii") for x in data[0:-1] if x[0] < 127 ])

	def __init__(self,fileID):
		temp = fileID.read(struct.calcsize(self.binary_format))
		data = struct.unpack(self.binary_format,temp)
		self.meshname = "".join([str(x, "ascii") for x in data[0:64] if x[0] < 127 ]) #Name of Mesh every Char is a  String on his own
		self.framecount    = data[64]      #Framecount = Number of Animationsteps
		self.vertexcount   = data[65]      #Number of Vertices in each Frame
		self.indexcount    = data[66]      #Number of Indices in Mesh (Triangles = Indexcount/3)
		self.diffusecolor  = data[67:70]   #RGB diffuse color
		self.specularcolor = data[70:73]   #RGB specular color (unused)
		self.specularpower = data[73]      #Specular power (unused)
		self.opacity       = data[74]      #Opacity
		self.properties    = data[75]      #Property flags
		self.textures      = data[76]      #Texture flags

		self.customalpha = bool(self.properties & 1)
		self.istwosided  = bool(self.properties & 2)
		self.noselect    = bool(self.properties & 4)
		self.glow    = bool(self.properties & 8)

		self.hastexture = False
		self.diffusetexture  = None
		self.speculartexture = None
		self.normaltexture   = None
		if self.textures:						#PropertyBit is Mesh Textured ?
			if self.textures & 1:  # diffuse
				self.diffusetexture = self._readtexname(fileID)
			if self.textures & 2:  # specular
				self.speculartexture = self._readtexname(fileID)
			if self.textures & 4:  # normal
				self.normaltexture = self._readtexname(fileID)

			self.hastexture = True
			# read all texture slots, otherwise it's read as data
			tex = self.textures >> 3
			while tex:
				tex &= tex - 1 # set rightmost 1-bit to 0
				# discard texture name, as we don't know what to do with it
				fileID.seek(struct.calcsize(self.texname_format))
				print("warning: ignored texture in undefined texture slot")

class G3DMeshdataV3:											   #Calculate and read the Mesh Datapack
	def __init__(self,fileID,header):
		#Calculation of the Meshdatasize to load because its variable
		#Animationframes * Vertices per Animation * 3 (Each Point are 3 Float X Y Z Coordinates)
		vertex_format = "<%if" % int(header.framecount * header.vertexcount * 3)
		#The same for Normals
		normals_format = "<%if" % int(header.normalframecount * header.vertexcount * 3)
		#Same here but Textures are 2D so only 2 Floats needed for Position inside Texture Bitmap
		texturecoords_format = "<%if" % int(header.texturecoordframecount * header.vertexcount * 2)
		#Colors in format RGBA
		colors_format = "<%if" % int(header.colorframecount * 4)
		#Indices
		indices_format = "<%iI" % int(header.indexcount)
		#Load the Meshdata as calculated above
		self.vertices = struct.unpack(vertex_format,fileID.read(struct.calcsize(vertex_format)))
		self.normals = struct.unpack(normals_format,fileID.read(struct.calcsize(normals_format)))
		self.texturecoords = struct.unpack(texturecoords_format,fileID.read(struct.calcsize(texturecoords_format)))
		self.colors = struct.unpack(colors_format,fileID.read(struct.calcsize(colors_format)))
		self.indices = struct.unpack(indices_format ,fileID.read(struct.calcsize(indices_format)))

class G3DMeshdataV4:											   #Calculate and read the Mesh Datapack
	def __init__(self,fileID,header):
		#Calculation of the Meshdatasize to load because its variable
		#Animationframes * Points (Vertex) per Animation * 3 (Each Point are 3 Float X Y Z Coordinates)
		vertex_format = "<%if" % int(header.framecount * header.vertexcount * 3)
		#The same for Normals
		normals_format = "<%if" % int(header.framecount * header.vertexcount * 3)
		#Same here but Textures are 2D so only 2 Floats needed for Position inside Texture Bitmap
		texturecoords_format = "<%if" % int(header.vertexcount * 2)
		#Indices
		indices_format = "<%iI" % int(header.indexcount)
		#Load the Meshdata as calculated above
		self.vertices = struct.unpack(vertex_format,fileID.read(struct.calcsize(vertex_format)))
		self.normals = struct.unpack(normals_format,fileID.read(struct.calcsize(normals_format)))
		if header.hastexture:
			self.texturecoords = struct.unpack(texturecoords_format,fileID.read(struct.calcsize(texturecoords_format)))
		self.indices = struct.unpack(indices_format ,fileID.read(struct.calcsize(indices_format)))

#Create a Mesh inside Blender
def createMesh(filename, header, data, toblender, operator):
	mesh = bpy.data.meshes.new(header.meshname)		#New Mesh
	meshobj = bpy.data.objects.new(header.meshname+'Object', mesh)	 #New Object for the new Mesh
	scene = bpy.context.scene
	scene.objects.link(meshobj)
	scene.update()
	uvcoords = []
	img_diffuse  = None
	img_specular = None
	img_normal   = None
	if header.hastexture:												  #Load Texture when assigned
		try:
			texturefile = dirname(abspath(filename)) + os.sep +	header.diffusetexture
			img_diffuse = bpy.data.images.load(texturefile)
			for x in range(0,len(data.texturecoords),2): #Prepare the UV
				uvcoords.append([data.texturecoords[x],data.texturecoords[x+1]])
			
			if header.isv4:
				if header.speculartexture:
					texturefile = dirname(abspath(filename)) + os.sep +	header.speculartexture
					img_specular = bpy.data.images.load(texturefile)
				if header.normaltexture:
					texturefile = dirname(abspath(filename)) + os.sep +	header.normaltexture
					img_normal = bpy.data.images.load(texturefile)
		except:
			import traceback
			traceback.print_exc()
			
			header.hastexture = False
			operator.report({'WARNING'}, "Couldn't load texture. See console for details.")
	
	vertsCO = []
	vertsNormal = []
	for x in range(0,header.vertexcount*3,3):	   #Get the Vertices and Normals into empty Mesh
		vertsCO.extend([(data.vertices[x],data.vertices[x+1],data.vertices[x+2])])
		vertsNormal.extend([(data.normals[x],data.normals[x+1],data.normals[x+2])])
		#vertsCO.extend([(data.vertices[x+(header.framecount-1)*header.vertexcount*3],data.vertices[x+(header.framecount-1)*header.vertexcount*3+1],data.vertices[x+(header.framecount-1)*header.vertexcount*3+2])])
		#vertsNormal.extend([(data.normals[x+(header.framecount-1)*header.vertexcount*3],data.normals[x+(header.framecount-1)*header.vertexcount*3+1],data.normals[x+(header.framecount-1)*header.vertexcount*3+2])])
	mesh.vertices.add(len(vertsCO))
	mesh.vertices.foreach_set("co", unpack_list(vertsCO))
	mesh.vertices.foreach_set("normal", unpack_list(vertsNormal))
	
	faces = []
	faceuv = []
	for i in range(0,len(data.indices),3):			  #Build Faces into Mesh
		faces.extend([data.indices[i], data.indices[i+1], data.indices[i+2], 0])
		if header.hastexture:	   
			uv = []
			u0 = uvcoords[data.indices[i]][0]
			v0 = uvcoords[data.indices[i]][1]
			uv.append([u0,v0])
			u1 = uvcoords[data.indices[i+1]][0]
			v1 = uvcoords[data.indices[i+1]][1]
			uv.append([u1,v1])
			u2 = uvcoords[data.indices[i+2]][0]
			v2 = uvcoords[data.indices[i+2]][1]
			uv.append([u2,v2])
			faceuv.append([uv,0,0,0])
		else:
			uv = []
			uv.append([0,0])
			uv.append([0,0])
			uv.append([0,0])
			faceuv.append([uv,0,0,0])
	mesh.tessfaces.add(len(faces)//4)
	mesh.tessfaces.foreach_set("vertices_raw", faces)
	mesh.tessfaces.foreach_set("use_smooth", [True] * len(mesh.tessfaces))
	mesh.g3d_customColor = header.customalpha
	mesh.show_double_sided = header.istwosided
	if header.isv4:
		mesh.g3d_noSelect = header.noselect
		mesh.g3d_glow = header.glow
	else:
		mesh.g3d_noSelect = False
		mesh.glow = False
	mesh.g3d_fullyOpaque = False
	#===================================================================================================
	#Material Setup
	#===================================================================================================
	def addtexslot(matdata,index,name,img):
		texture = bpy.data.textures.new(name=name,type='IMAGE')
		texture.image = img
		slot = matdata.texture_slots.create(index)
		slot.texture = texture
		slot.texture_coords = 'UV'

	if header.hastexture:	   
		materialname = "pskmat"
		materials = []
		matdata = bpy.data.materials.new(materialname + '1')

		addtexslot(matdata, 0, 'diffusetexture', img_diffuse)
		if img_specular:
			addtexslot(matdata, 1, 'speculartexture', img_specular)
		if img_normal:
			addtexslot(matdata, 2, 'normaltexture', img_normal)

		if header.isv4:
			matdata.diffuse_color = (header.diffusecolor[0], header.diffusecolor[1],header.diffusecolor[2])
			matdata.alpha = header.opacity
			matdata.specular_color = (header.specularcolor[0], header.specularcolor[1],header.specularcolor[2])
		materials.append(matdata)

		for material in materials:
			#add material to the mesh list of materials
			mesh.materials.append(material)

		countm = 0
		psktexname="psk" + str(countm)
		mesh.tessface_uv_textures.new(name=psktexname)
		if (len(faceuv) > 0):
			for countm in range(len(mesh.tessface_uv_textures)):
				uvtex = mesh.tessface_uv_textures[countm] #add one uv texture
				for i, face in enumerate(mesh.tessfaces):
					blender_tface = uvtex.data[i] #face
					mfaceuv = faceuv[i]
					if countm == faceuv[i][1]:
						face.material_index = faceuv[i][1]
						blender_tface.uv1 = mfaceuv[0][0] #uv = (0,0)
						blender_tface.uv2 = mfaceuv[0][1] #uv = (0,0)
						blender_tface.uv3 = mfaceuv[0][2] #uv = (0,0)
						blender_tface.image = img_diffuse
					else:
						blender_tface.uv1 = [0,0]
						blender_tface.uv2 = [0,0]
						blender_tface.uv3 = [0,0]
	imported.append(meshobj)			#Add to Imported Objects
	sk = meshobj.shape_key_add()
	for x in range(1,header.framecount):	#Put in Vertex Positions for Keyanimation
		sk = meshobj.shape_key_add()
		for i in range(0,header.vertexcount*3,3):
			sk.data[i//3].co[0]= data.vertices[x*header.vertexcount*3 + i]
			sk.data[i//3].co[1]= data.vertices[x*header.vertexcount*3 + i +1]
			sk.data[i//3].co[2]= data.vertices[x*header.vertexcount*3 + i +2]

	# activate one shapekey per frame
	for i in range(1,header.framecount):
		shape = mesh.shape_keys.key_blocks[i]
		shape.value = 0.0
		shape.keyframe_insert("value", frame=i)
		shape.value = 1.0
		shape.keyframe_insert("value", frame=(i+1))
		shape.value = 0.0
		shape.keyframe_insert("value", frame=(i+2))

	meshobj.active_shape_key_index = 0

	if toblender:
		# rotate from glest to blender orientation
		#mesh.transform( Matrix( ((1,0,0,0),(0,0,-1,0),(0,1,0,0),(0,0,0,1)) ) )
		# doesn't work, maybe because of shape keys
		# use object transformation instead
		meshobj.rotation_euler = (radians(90), 0, 0)

	# update polygon structures from tessfaces
	mesh.update()
	mesh.update_tag()

	# remove duplicates
	bm = bmesh.new()
	bm.from_mesh(mesh)
	bmesh.ops.remove_doubles(bm, verts=bm.verts, dist=0.0001)
	bm.to_mesh(mesh)
	bm.free()

	return
###########################################################################
# Import
###########################################################################
def G3DLoader(filepath, toblender, operator):			#Main Import Routine
	global imported, sceneID
	print ("\nNow Importing File: " + filepath)
	fileID = open(filepath,"rb")
	header = G3DHeader(fileID)
	print ("\nHeader ID         : " + header.id)
	print ("Version           : " + str(header.version))
	if header.id != "G3D":
		print ("ERROR: This is Not a G3D Model File")
		operator.report({'ERROR'}, "This is Not a G3D Model File")
		fileID.close
		return
	if header.version not in (3, 4):
		print ("ERROR: The Version of this G3D File is not Supported")
		operator.report({'ERROR'}, "The Version of this G3D File is not Supported")
		fileID.close
		return
	#in_editmode = Blender.Window.EditMode()			 #Must leave Editmode when active
	#if in_editmode: Blender.Window.EditMode(0)
	sceneID = bpy.context.scene						  #Get active Scene
	#scenecontext=sceneID.getRenderingContext()		  #To Access the Start/Endframe its so hidden i searched till i got angry :-)
	basename=os.path.basename(filepath).split('.')[0]   #Generate the Base Filename without Path + extension
	imported = []
	maxframe=0
	if header.version == 3:
		modelheader = G3DModelHeaderv3(fileID)
		print ("Number of Meshes  : " + str(modelheader.meshcount))
		for x in range(modelheader.meshcount):
			meshheader = G3DMeshHeaderv3(fileID)
			meshheader.isv4 = False	
			print ("\nMesh Number         : " + str(x+1))
			print ("framecount            : " + str(meshheader.framecount))
			print ("normalframecount      : " + str(meshheader.normalframecount))
			print ("texturecoordframecount: " + str(meshheader.texturecoordframecount))
			print ("colorframecount       : " + str(meshheader.colorframecount))
			print ("pointcount            : " + str(meshheader.vertexcount))
			print ("indexcount            : " + str(meshheader.indexcount))
			print ("texturename           : " + str(meshheader.diffusetexture))
			print ("hastexture            : " + str(meshheader.hastexture))
			print ("istwosided            : " + str(meshheader.istwosided))
			print ("customalpha           : " + str(meshheader.customalpha))
			meshheader.meshname = basename+str(x+1)	 #Generate Meshname because V3 has none
			if meshheader.framecount > maxframe: maxframe = meshheader.framecount #Evaluate the maximal animationsteps
			meshdata = G3DMeshdataV3(fileID,meshheader)
			createMesh(filepath, meshheader, meshdata, toblender, operator)
		fileID.close
		bpy.context.scene.frame_start=1
		bpy.context.scene.frame_end=maxframe
		bpy.context.scene.frame_current=1
		anchor = bpy.data.objects.new('Empty', None)
		anchor.select = True
		bpy.context.scene.objects.link(anchor)
		for ob in imported:
				ob.parent = anchor
		bpy.context.scene.update()
		return
	if header.version == 4:
		modelheader = G3DModelHeaderv4(fileID)
		print ("Number of Meshes  : " + str(modelheader.meshcount))
		for x in range(modelheader.meshcount):
			meshheader = G3DMeshHeaderv4(fileID)
			meshheader.isv4 = True	  
			print ("\nMesh Number   : " + str(x+1))
			print ("meshname        : " + str(meshheader.meshname))
			print ("framecount      : " + str(meshheader.framecount))
			print ("vertexcount     : " + str(meshheader.vertexcount))
			print ("indexcount      : " + str(meshheader.indexcount))
			print ("diffusecolor    : %1.6f %1.6f %1.6f" %meshheader.diffusecolor)
			print ("specularcolor   : %1.6f %1.6f %1.6f" %meshheader.specularcolor)
			print ("specularpower   : %1.6f" %meshheader.specularpower)
			print ("opacity         : %1.6f" %meshheader.opacity)
			print ("properties      : " + str(meshheader.properties))
			print ("textures        : " + str(meshheader.textures))
			print ("texturename     : " + str(meshheader.diffusetexture))
			if len(meshheader.meshname) ==0:	#When no Meshname in File Generate one
					meshheader.meshname = basename+str(x+1)
			if meshheader.framecount > maxframe: maxframe = meshheader.framecount #Evaluate the maximal animationsteps
			meshdata = G3DMeshdataV4(fileID,meshheader)
			createMesh(filepath, meshheader, meshdata, toblender, operator)
		fileID.close
		
		bpy.context.scene.frame_start=1
		bpy.context.scene.frame_end=maxframe
		bpy.context.scene.frame_current=1
		anchor = bpy.data.objects.new('Empty', None)
		anchor.select = True
		bpy.context.scene.objects.link(anchor)
		for ob in imported:
				ob.parent = anchor
		bpy.context.scene.update()
		print ("Created a empty Object as 'Grip' where all imported Objects are parented to")
		print ("To move the complete Meshes only select this empty Object and move it")
		print ("All Done, have a good Day :-)\n\n")
		return

def G3DSaver(filepath, context, toglest, operator):
	print ("\nNow Exporting File: " + filepath)

	objs = context.selected_objects
	if len(objs) == 0:
		objs = bpy.data.objects

	#get real meshcount as len(bpy.data.meshes) holds also old meshes
	meshCount = 0
	for obj in objs:
		if obj.type == 'MESH':
			meshCount += 1
			if obj.mode != 'OBJECT': # we want to be in object mode
				print("ERROR: mesh not in object mode")
				operator.report({'ERROR'}, "mesh not in object mode")
				return -1

	if meshCount == 0:
		print("ERROR: no meshes found")
		operator.report({'ERROR'}, "no meshes found")
		return -1

	fileID = open(filepath,"wb")
	# G3DHeader v4
	fileID.write(struct.pack("<3cB", b'G', b'3', b'D', 4))
	# G3DModelHeaderv4
	fileID.write(struct.pack("<HB", meshCount, 0))
	# meshes
	#for mesh in bpy.data.meshes:
	for obj in objs:
		if obj.type != 'MESH':
			continue
		mesh = obj.data.copy()
		diffuseColor = [1.0, 1.0, 1.0]
		specularColor = [0.9, 0.9, 0.9]
		opacity = 1.0
		textures = 0
		if len(mesh.materials) > 0:
			# we have a texture, hopefully
			material = mesh.materials[0]
			slot = material.texture_slots[0]
			# only look for other textures when we have diffuse
			if slot and slot.texture.type=='IMAGE' and len(mesh.uv_textures)>0:
				diffuseColor = material.diffuse_color
				specularColor = material.specular_color
				opacity = material.alpha
				textures = 1
				texnames = []
				texnames.append(bpy.path.basename(slot.texture.image.filepath))
				# specular and normal
				for i in range(1, 3):
					slot = material.texture_slots[i]
					if slot and slot.texture.type=='IMAGE':
						texnames.append(bpy.path.basename(slot.texture.image.filepath))
						textures |= 1 << i
					
			else:
				print("WARNING: first texture slot in first material isn't of type IMAGE or it's not unwrapped, texture ignored")
				operator.report({'WARNING'}, "first texture slot in first material isn't of type IMAGE or it's not unwrapped, texture ignored")
				#continue without texture

		meshname = mesh.name
		frameCount = context.scene.frame_end - context.scene.frame_start +1
		realFaceCount = 0 # real face count (triangles)
		indices=[]        # list of indices
		newverts=[]       # list of vertex indices which need to be duplicated
		uvlist = []       # list of texcoords
		mesh.update(calc_tessface=True) # tesselate n-polygons to triangles & quads
		if textures:
			uvtex = mesh.tessface_uv_textures[0]
			uvlist[:] = [[0]*2 for i in range(len(mesh.vertices))]
			# blender allows to have multiple texcoords per vertex,
			# in g3d format every vertex can only have one texcoord
			# -> duplicate vertex
			# the dictionary/map vdict collects all the stuff
			# index to "unique" vertices from blender
			#   -> tuple( list of texcoords, list of indices to the duplicated vertex )
			vdict = dict()
			nextIndex = len(mesh.vertices)
			for face in mesh.tessfaces:
				# when a vertex is duplicated it gets a new index, so the
				# triple of indices describing the face is different too
				faceindices = []
				realFaceCount += 1
				uvdata = uvtex.data[face.index]
				for i in range(3):
					#closure, got rid of copy&paste, still looking weird
					def getTexCoords():
						nonlocal nextIndex, vdict, uvlist, newverts
						vindex = face.vertices[i]
						if vindex not in vdict: # new vertex -> add it
							vdict[vindex] = [uvdata.uv[i]], [vindex]
							uvlist[vindex] = uvdata.uv[i] # that's a (s,t)-pair
						else:
							found = False
							idx = 0
							for ele in vdict[vindex][0]:
								if uvdata.uv[i][0] == ele[0] and uvdata.uv[i][1] == ele[1]:
									found = True
									break
								idx += 1
							if found: # same vertex and texcoord before
								# it could be a different index now, the index of a new
								# duplicated vertex
								#vindex = vdict[vindex][1][ vdict[vindex][0].index(uvdata.uv[i]) ]
								vindex = vdict[vindex][1][idx]
							else: # same vertex as before but with different texcoord -> duplicate
								vdict[vindex][0].append(uvdata.uv[i])
								vdict[vindex][1].append(nextIndex)

								# duplicate vertex because it takes part in different faces
								# with different texcoords
								newverts.append(vindex)
								uvlist.append(uvdata.uv[i])
								# new index for the duplicated vertex
								vindex = nextIndex
								nextIndex += 1

						faceindices.append(vindex)
					getTexCoords()
				indices.extend(faceindices)

				if len(face.vertices) == 4:
					faceindices = []
					realFaceCount += 1
					for i in [0,2,3]:
						getTexCoords()
					indices.extend(faceindices)
		else:
			for face in mesh.tessfaces:
				realFaceCount += 1
				indices.extend(face.vertices[0:3])
				if len(face.vertices) == 4:
					realFaceCount += 1
					# new face because quad got split
					indices.append(face.vertices[0])
					indices.append(face.vertices[2])
					indices.append(face.vertices[3])


		# abort when no triangles as it crashs g3dviewer
		if realFaceCount == 0:
			print("ERROR: no triangles found")
			operator.report({'ERROR'}, "no triangles found")
			fileID.close()
			return -1
		indexCount = realFaceCount * 3
		vertexCount = len(mesh.vertices) + len(newverts)
		specularPower = 9.999999  # unused, same as old exporter
		properties = 0
		if mesh.g3d_customColor:
			properties |= 1
		if mesh.show_double_sided:
			properties |= 2
		if mesh.g3d_noSelect:
			properties |= 4
		if mesh.g3d_glow:
			properties |= 8

		#MeshData
		vertices = []
		normals = []
		fcurrent = context.scene.frame_current
		for i in range(context.scene.frame_start, context.scene.frame_end+1):
			context.scene.frame_set(i)
			#FIXME: not sure what's better: PREVIEW or RENDER settings
			m = obj.to_mesh(context.scene, True, 'RENDER')
			m.transform(obj.matrix_world)  # apply object-mode transformations

			if toglest:
				# rotate from blender to glest orientation
				m.transform( Matrix( ((1,0,0,0),(0,0,1,0),(0,-1,0,0),(0,0,0,1)) ) )
				# transform normals too
				m.calc_normals()

			for vertex in m.vertices:
				vertices.extend(vertex.co)
				normals.extend(vertex.normal)

			# duplicate vertices and corresponding normals, for every frame
			for nv in newverts:
				vertices.extend(m.vertices[nv].co)
				normals.extend(m.vertices[nv].normal)

		context.scene.frame_set(fcurrent)

		if mesh.g3d_fullyOpaque:
			opacity = 1.0

		# MeshHeader
		fileID.write(struct.pack("<64s3I8f2I",
			bytes(meshname, "ascii"),
			frameCount, vertexCount, indexCount,
			diffuseColor[0], diffuseColor[1], diffuseColor[2],
			specularColor[0], specularColor[1], specularColor[2],
			specularPower, opacity,
			properties, textures
		))
		#Texture names
		if textures: # only when we have textures
			for i in range(len(texnames)):
				fileID.write(struct.pack("<64s", bytes(texnames[i], "ascii")))

		# see G3DMeshdataV4
		vertex_format = "<%if" % int(frameCount * vertexCount * 3)
		normals_format = "<%if" % int(frameCount * vertexCount * 3)
		texturecoords_format = "<%if" % int(vertexCount * 2)
		indices_format = "<%iI" % int(indexCount)

		fileID.write(struct.pack(vertex_format, *vertices))
		fileID.write(struct.pack(normals_format, *normals))

		# texcoords
		if textures: # only when we have textures
			texcoords = []
			for uv in uvlist:
				texcoords.extend(uv)
			fileID.write(struct.pack(texturecoords_format, *texcoords))

		fileID.write(struct.pack(indices_format, *indices))

	fileID.close()
	return 0


#---=== Register ===
class G3DPanel(bpy.types.Panel):
	#bl_idname = "OBJECT_PT_G3DPanel"
	bl_label = "G3D properties"
	bl_space_type = 'PROPERTIES'
	bl_region_type = 'WINDOW'
	bl_context = "data"

	@classmethod
	def poll(cls, context):
		return (context.object is not None and context.object.type == 'MESH')

	def draw(self, context):
		self.layout.prop(context.object.data, "g3d_customColor")
		self.layout.prop(context.object.data, "show_double_sided", text="double sided")
		self.layout.prop(context.object.data, "g3d_noSelect")
		self.layout.prop(context.object.data, "g3d_fullyOpaque")
		self.layout.prop(context.object.data, "g3d_glow")

class ImportG3D(bpy.types.Operator, ImportHelper):
	'''Load a G3D file'''
	bl_idname = "importg3d.g3d"
	bl_label = "Import G3D"

	filename_ext = ".g3d"
	filter_glob = StringProperty(default="*.g3d", options={'HIDDEN'})

	toblender = bpy.props.BoolProperty(
				name="rotate to Blender orientation",
				description="Rotate meshes from Glest to Blender orientation",
				default=True)

	def execute(self, context):
		try:
			G3DLoader(self.filepath, self.toblender, self)
		except:
			import traceback
			traceback.print_exc()

			return {'CANCELLED'}

		return {'FINISHED'}

class ExportG3D(bpy.types.Operator, ExportHelper):
	'''Save a G3D file'''
	bl_idname = "exportg3d.g3d"
	bl_label = "Export G3D"

	filename_ext = ".g3d"
	filter_glob = StringProperty(default="*.g3d", options={'HIDDEN'})

	#export options
	showg3d = bpy.props.BoolProperty(
				name="show G3D afterwards",
				description=("Run g3dviewer to show G3D after export. "
							"g3dviewer needs to be in the scripts directory, "
							"otherwise the associated program of .g3d is run."),
				default=False)
	toglest = bpy.props.BoolProperty(
				name="rotate to glest orientation",
				description="Rotate meshes from Blender to Glest orientation",
				default=True)

	def execute(self, context):
		try:
			res = G3DSaver(self.filepath, context, self.toglest, self)
			if res==0 and self.showg3d:
				print("opening g3dviewer with " + self.filepath)
				scriptsdir = bpy.utils.script_path_user()
				dname = os.path.dirname(self.filepath)
				found = False
				for f in os.listdir(scriptsdir):
					if "g3dviewer" in f:
						f = os.path.join(scriptsdir, f)
						if os.path.isfile(f) and os.access(f, os.X_OK):
							cmd = [f, self.filepath]
							print(cmd)
							subprocess.Popen(cmd, cwd=dname)
							found = True

				# try default associated program
				if not found:
					if os.name == 'posix':
						# xdg-open is only a shell script which delegates the job to a
						# desktop specific program, e.g. if DE=kde than kde-open
						# needs DE environment variable set, otherwise it just throws it
						# at the browser, which is not very helpful
						print("running xdg-open "+self.filepath)
						subprocess.Popen(['xdg-open', self.filepath], cwd=dname)
					elif os.name == 'mac':
						subprocess.Popen(['open', self.filepath], cwd=dname)
					elif os.name == 'nt':
						#os.startfile(self.filepath) # no way to change dir
						subprocess.Popen(['cmd', '/C', 'start', self.filepath], cwd=dname)

		except:
			import traceback
			traceback.print_exc()

			return {'CANCELLED'}

		return {'FINISHED'}

def menu_func_import(self, context):
	self.layout.operator(ImportG3D.bl_idname, text="Glest 3D File (.g3d)")

def menu_func_export(self, context):
	self.layout.operator(ExportG3D.bl_idname, text="Glest 3D File (.g3d)")

def register():
	# custom mesh properties
	bpy.types.Mesh.g3d_customColor = bpy.props.BoolProperty(
			name="team color",
			description="replace alpha channel of texture with team color")
	bpy.types.Mesh.g3d_noSelect = bpy.props.BoolProperty(
			name="non-selectable",
			description="click on mesh doesn't select unit")
	bpy.types.Mesh.g3d_fullyOpaque = bpy.props.BoolProperty(
			name="fully opaque",
			description="sets opacity to 1.0, ignoring what's set in materials")
	bpy.types.Mesh.g3d_glow = bpy.props.BoolProperty(
			name="glow",
			description="let objects glow like particles")

	bpy.utils.register_module(__name__)

	bpy.types.INFO_MT_file_import.append(menu_func_import)
	bpy.types.INFO_MT_file_export.append(menu_func_export)


def unregister():
	bpy.utils.unregister_module(__name__)

	bpy.types.INFO_MT_file_import.remove(menu_func_import)
	bpy.types.INFO_MT_file_export.remove(menu_func_export)

if __name__ == '__main__':
	register()
#	main()

	#for obj in bpy.data.objects:
	#	if obj.type == 'MESH':
	#		obj.select = True
	#		bpy.ops.object.delete()
	#G3DLoader("import.g3d", True, None)

	#for obj in bpy.context.selected_objects:
	#	obj.select = False  # deselect everything, so we get it all
	#G3DSaver("test.g3d", bpy.context)




#!/bin/env python
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Usage: symbolstore.py <params> <dump_syms path> <symbol store path>
#                                <debug info files or dirs>
#   Runs dump_syms on each debug info file specified on the command line,
#   then places the resulting symbol file in the proper directory
#   structure in the symbol store path.  Accepts multiple files
#   on the command line, so can be called as part of a pipe using
#   find <dir> | xargs symbolstore.pl <dump_syms> <storepath>
#   But really, you might just want to pass it <dir>.
#
#   Parameters accepted:
#     -c           : Copy debug info files to the same directory structure
#                    as sym files
#     -a "<archs>" : Run dump_syms -a <arch> for each space separated
#                    cpu architecture in <archs> (only on OS X)
#     -s <srcdir>  : Use <srcdir> as the top source directory to
#                    generate relative filenames.

import sys
import platform
import os
import re
import shutil
import textwrap
import fnmatch
import subprocess
import urlparse
import multiprocessing
import collections
from optparse import OptionParser
from xml.dom.minidom import parse

# Utility classes

class VCSFileInfo:
    """ A base class for version-controlled file information. Ensures that the
        following attributes are generated only once (successfully):

            self.root
            self.clean_root
            self.revision
            self.filename

        The attributes are generated by a single call to the GetRoot,
        GetRevision, and GetFilename methods. Those methods are explicitly not
        implemented here and must be implemented in derived classes. """

    def __init__(self, file):
        if not file:
            raise ValueError
        self.file = file

    def __getattr__(self, name):
        """ __getattr__ is only called for attributes that are not set on self,
            so setting self.[attr] will prevent future calls to the GetRoot,
            GetRevision, and GetFilename methods. We don't set the values on
            failure on the off chance that a future call might succeed. """

        if name == "root":
            root = self.GetRoot()
            if root:
                self.root = root
            return root

        elif name == "clean_root":
            clean_root = self.GetCleanRoot()
            if clean_root:
                self.clean_root = clean_root
            return clean_root

        elif name == "revision":
            revision = self.GetRevision()
            if revision:
                self.revision = revision
            return revision

        elif name == "filename":
            filename = self.GetFilename()
            if filename:
                self.filename = filename
            return filename

        raise AttributeError

    def GetRoot(self):
        """ This method should return the unmodified root for the file or 'None'
            on failure. """
        raise NotImplementedError

    def GetCleanRoot(self):
        """ This method should return the repository root for the file or 'None'
            on failure. """
        raise NotImplementedErrors

    def GetRevision(self):
        """ This method should return the revision number for the file or 'None'
            on failure. """
        raise NotImplementedError

    def GetFilename(self):
        """ This method should return the repository-specific filename for the
            file or 'None' on failure. """
        raise NotImplementedError


# This regex separates protocol and optional username/password from a url.
# For instance, all the following urls will be transformed into
# 'foo.com/bar':
#
#   http://foo.com/bar
#   svn+ssh://user@foo.com/bar
#   svn+ssh://user:pass@foo.com/bar
#
rootRegex = re.compile(r'^\S+?:/+(?:[^\s/]*@)?(\S+)$')

def read_output(*args):
    (stdout, _) = subprocess.Popen(args=args, stdout=subprocess.PIPE).communicate()
    return stdout.rstrip()

class HGRepoInfo:
    def __init__(self, path):
        self.path = path
        rev = read_output('hg', '-R', path,
                          'parent', '--template={node|short}')
        # Look for the default hg path.  If SRVSRV_ROOT is set, we
        # don't bother asking hg.
        hg_root = os.environ.get("SRCSRV_ROOT")
        if hg_root:
            root = hg_root
        else:
            root = read_output('hg', '-R', path,
                               'showconfig', 'paths.default')
            if not root:
                print >> sys.stderr, "Failed to get HG Repo for %s" % path
        cleanroot = None
        if root:
            match = rootRegex.match(root)
            if match:
                cleanroot = match.group(1)
                if cleanroot.endswith('/'):
                    cleanroot = cleanroot[:-1]
        if cleanroot is None:
            print >> sys.stderr, textwrap.dedent("""\
                Could not determine repo info for %s.  This is either not a clone of the web-based
                repository, or you have not specified SRCSRV_ROOT, or the clone is corrupt.""") % path
            sys.exit(1)
        self.rev = rev
        self.cleanroot = cleanroot

    def GetFileInfo(self, file):
        return HGFileInfo(file, self)

class HGFileInfo(VCSFileInfo):
    def __init__(self, file, repo):
        VCSFileInfo.__init__(self, file)
        self.repo = repo
        self.file = os.path.relpath(file, repo.path)

    def GetRoot(self):
        return self.repo.path

    def GetCleanRoot(self):
        return self.repo.cleanroot

    def GetRevision(self):
        return self.repo.rev

    def GetFilename(self):
        if self.revision and self.clean_root:
            return "hg:%s:%s:%s" % (self.clean_root, self.file, self.revision)
        return self.file

class GitRepoInfo:
    """
    Info about a local git repository. Does not currently
    support discovering info about a git clone, the info must be
    provided out-of-band.
    """
    def __init__(self, path, rev, root):
        self.path = path
        cleanroot = None
        if root:
            match = rootRegex.match(root)
            if match:
                cleanroot = match.group(1)
                if cleanroot.endswith('/'):
                    cleanroot = cleanroot[:-1]
        if cleanroot is None:
            print >> sys.stderr, textwrap.dedent("""\
                Could not determine repo info for %s (%s).  This is either not a clone of a web-based
                repository, or you have not specified SRCSRV_ROOT, or the clone is corrupt.""") % (path, root)
            sys.exit(1)
        self.rev = rev
        self.cleanroot = cleanroot

    def GetFileInfo(self, file):
        return GitFileInfo(file, self)

class GitFileInfo(VCSFileInfo):
    def __init__(self, file, repo):
        VCSFileInfo.__init__(self, file)
        self.repo = repo
        self.file = os.path.relpath(file, repo.path)

    def GetRoot(self):
        return self.repo.path

    def GetCleanRoot(self):
        return self.repo.cleanroot

    def GetRevision(self):
        return self.repo.rev

    def GetFilename(self):
        if self.revision and self.clean_root:
            return "git:%s:%s:%s" % (self.clean_root, self.file, self.revision)
        return self.file

# Utility functions

# A cache of repo info for each srcdir.
srcdirRepoInfo = {}

# A cache of files for which VCS info has already been determined. Used to
# prevent extra filesystem activity or process launching.
vcsFileInfoCache = {}

def IsInDir(file, dir):
    # the lower() is to handle win32+vc8, where
    # the source filenames come out all lowercase,
    # but the srcdir can be mixed case
    return os.path.abspath(file).lower().startswith(os.path.abspath(dir).lower())

def GetVCSFilenameFromSrcdir(file, srcdir):
    if srcdir not in srcdirRepoInfo:
        # Not in cache, so find it adnd cache it
        if os.path.isdir(os.path.join(srcdir, '.hg')):
            srcdirRepoInfo[srcdir] = HGRepoInfo(srcdir)
        else:
            # Unknown VCS or file is not in a repo.
            return None
    return srcdirRepoInfo[srcdir].GetFileInfo(file)

def GetVCSFilename(file, srcdirs):
    """Given a full path to a file, and the top source directory,
    look for version control information about this file, and return
    a tuple containing
    1) a specially formatted filename that contains the VCS type,
    VCS location, relative filename, and revision number, formatted like:
    vcs:vcs location:filename:revision
    For example:
    cvs:cvs.mozilla.org/cvsroot:mozilla/browser/app/nsBrowserApp.cpp:1.36
    2) the unmodified root information if it exists"""
    (path, filename) = os.path.split(file)
    if path == '' or filename == '':
        return (file, None)

    fileInfo = None
    root = ''
    if file in vcsFileInfoCache:
        # Already cached this info, use it.
        fileInfo = vcsFileInfoCache[file]
    else:
        for srcdir in srcdirs:
            if not IsInDir(file, srcdir):
                continue
            fileInfo = GetVCSFilenameFromSrcdir(file, srcdir)
            if fileInfo:
                vcsFileInfoCache[file] = fileInfo
                break

    if fileInfo:
        file = fileInfo.filename
        root = fileInfo.root

    # we want forward slashes on win32 paths
    return (file.replace("\\", "/"), root)

def GetPlatformSpecificDumper(**kwargs):
    """This function simply returns a instance of a subclass of Dumper
    that is appropriate for the current platform."""
    # Python 2.5 has a bug where platform.system() returns 'Microsoft'.
    # Remove this when we no longer support Python 2.5.
    return {'Windows': Dumper_Win32,
            'Microsoft': Dumper_Win32,
            'Linux': Dumper_Linux,
            'Sunos5': Dumper_Solaris,
            'Darwin': Dumper_Mac}[platform.system()](**kwargs)

def SourceIndex(fileStream, outputPath, vcs_root):
    """Takes a list of files, writes info to a data block in a .stream file"""
    # Creates a .pdb.stream file in the mozilla\objdir to be used for source indexing
    # Create the srcsrv data block that indexes the pdb file
    result = True
    pdbStreamFile = open(outputPath, "w")
    pdbStreamFile.write('''SRCSRV: ini ------------------------------------------------\r\nVERSION=2\r\nINDEXVERSION=2\r\nVERCTRL=http\r\nSRCSRV: variables ------------------------------------------\r\nHGSERVER=''')
    pdbStreamFile.write(vcs_root)
    pdbStreamFile.write('''\r\nSRCSRVVERCTRL=http\r\nHTTP_EXTRACT_TARGET=%hgserver%/raw-file/%var3%/%var2%\r\nSRCSRVTRG=%http_extract_target%\r\nSRCSRV: source files ---------------------------------------\r\n''')
    pdbStreamFile.write(fileStream) # can't do string interpolation because the source server also uses this and so there are % in the above
    pdbStreamFile.write("SRCSRV: end ------------------------------------------------\r\n\n")
    pdbStreamFile.close()
    return result

def WorkerInitializer(cls, lock):
    """Windows worker processes won't have run GlobalInit, and due to a lack of fork(),
    won't inherit the class variables from the parent. The only one they need is the lock,
    so we run an initializer to set it. Redundant but harmless on other platforms."""
    cls.lock = lock

def StartProcessFilesWork(dumper, files, arch_num, arch, vcs_root, after, after_arg):
    """multiprocessing can't handle methods as Process targets, so we define
    a simple wrapper function around the work method."""
    return dumper.ProcessFilesWork(files, arch_num, arch, vcs_root, after, after_arg)

class Dumper:
    """This class can dump symbols from a file with debug info, and
    store the output in a directory structure that is valid for use as
    a Breakpad symbol server.  Requires a path to a dump_syms binary--
    |dump_syms| and a directory to store symbols in--|symbol_path|.
    Optionally takes a list of processor architectures to process from
    each debug file--|archs|, the full path to the top source
    directory--|srcdir|, for generating relative source file names,
    and an option to copy debug info files alongside the dumped
    symbol files--|copy_debug|, mostly useful for creating a
    Microsoft Symbol Server from the resulting output.

    You don't want to use this directly if you intend to call
    ProcessDir.  Instead, call GetPlatformSpecificDumper to
    get an instance of a subclass.
 
    Processing is performed asynchronously via worker processes; in
    order to wait for processing to finish and cleanup correctly, you
    must call Finish after all Process/ProcessDir calls have been made.
    You must also call Dumper.GlobalInit before creating or using any
    instances."""
    def __init__(self, dump_syms, symbol_path,
                 archs=None,
                 srcdirs=[],
                 copy_debug=False,
                 vcsinfo=False,
                 srcsrv=False,
                 exclude=[],
                 repo_manifest=None):
        # popen likes absolute paths, at least on windows
        self.dump_syms = os.path.abspath(dump_syms)
        self.symbol_path = symbol_path
        if archs is None:
            # makes the loop logic simpler
            self.archs = ['']
        else:
            self.archs = ['-a %s' % a for a in archs.split()]
        self.srcdirs = [os.path.normpath(a) for a in srcdirs]
        self.copy_debug = copy_debug
        self.vcsinfo = vcsinfo
        self.srcsrv = srcsrv
        self.exclude = exclude[:]
        if repo_manifest:
            self.parse_repo_manifest(repo_manifest)

        # book-keeping to keep track of our jobs and the cleanup work per file tuple
        self.files_record = {}
        self.jobs_record = collections.defaultdict(int)

    @classmethod
    def GlobalInit(cls, module=multiprocessing):
        """Initialize the class globals for the multiprocessing setup; must
        be called before any Dumper instances are created and used. Test cases
        may pass in a different module to supply Manager and Pool objects,
        usually multiprocessing.dummy."""
        num_cpus = module.cpu_count()
        if num_cpus is None:
            # assume a dual core machine if we can't find out for some reason
            # probably better on single core anyway due to I/O constraints
            num_cpus = 2

        # have to create any locks etc before the pool
        cls.manager = module.Manager()
        cls.jobs_condition = Dumper.manager.Condition()
        cls.lock = Dumper.manager.RLock()
        cls.pool = module.Pool(num_cpus, WorkerInitializer, (cls, cls.lock))

    def JobStarted(self, file_key):
        """Increments the number of submitted jobs for the specified key file,
        defined as the original file we processed; note that a single key file
        can generate up to 1 + len(self.archs) jobs in the Mac case."""
        with Dumper.jobs_condition:
            self.jobs_record[file_key] += 1
            Dumper.jobs_condition.notify_all()

    def JobFinished(self, file_key):
        """Decrements the number of submitted jobs for the specified key file,
        defined as the original file we processed; once the count is back to 0,
        remove the entry from our record."""
        with Dumper.jobs_condition:
            self.jobs_record[file_key] -= 1

            if self.jobs_record[file_key] == 0:
                del self.jobs_record[file_key]

            Dumper.jobs_condition.notify_all()

    def output(self, dest, output_str):
        """Writes |output_str| to |dest|, holding |lock|;
        terminates with a newline."""
        with Dumper.lock:
            dest.write(output_str + "\n")
            dest.flush()

    def output_pid(self, dest, output_str):
        """Debugging output; prepends the pid to the string."""
        self.output(dest, "%d: %s" % (os.getpid(), output_str))

    def parse_repo_manifest(self, repo_manifest):
        """
        Parse an XML manifest of repository info as produced
        by the `repo manifest -r` command.
        """
        doc = parse(repo_manifest)
        if doc.firstChild.tagName != "manifest":
            return
        # First, get remotes.
        remotes = dict([(r.getAttribute("name"), r.getAttribute("fetch")) for r in doc.getElementsByTagName("remote")])
        # And default remote.
        default_remote = None
        if doc.getElementsByTagName("default"):
            default_remote = doc.getElementsByTagName("default")[0].getAttribute("remote")
        # Now get projects. Assume they're relative to repo_manifest.
        base_dir = os.path.abspath(os.path.dirname(repo_manifest))
        for proj in doc.getElementsByTagName("project"):
            # name is the repository URL relative to the remote path.
            name = proj.getAttribute("name")
            # path is the path on-disk, relative to the manifest file.
            path = proj.getAttribute("path")
            # revision is the changeset ID.
            rev = proj.getAttribute("revision")
            # remote is the base URL to use.
            remote = proj.getAttribute("remote")
            # remote defaults to the <default remote>.
            if not remote:
                remote = default_remote
            # path defaults to name.
            if not path:
                path = name
            if not (name and path and rev and remote):
                print "Skipping project %s" % proj.toxml()
                continue
            remote = remotes[remote]
            # Turn git URLs into http URLs so that urljoin works.
            if remote.startswith("git:"):
                remote = "http" + remote[3:]
            # Add this project to srcdirs.
            srcdir = os.path.join(base_dir, path)
            self.srcdirs.append(srcdir)
            # And cache its VCS file info. Currently all repos mentioned
            # in a repo manifest are assumed to be git.
            root = urlparse.urljoin(remote, name)
            srcdirRepoInfo[srcdir] = GitRepoInfo(srcdir, rev, root)

    # subclasses override this
    def ShouldProcess(self, file):
        return not any(fnmatch.fnmatch(os.path.basename(file), exclude) for exclude in self.exclude)

    # and can override this
    def ShouldSkipDir(self, dir):
        return False

    def RunFileCommand(self, file):
        """Utility function, returns the output of file(1)"""
        try:
            # we use -L to read the targets of symlinks,
            # and -b to print just the content, not the filename
            return os.popen("file -Lb " + file).read()
        except:
            return ""

    # This is a no-op except on Win32
    def FixFilenameCase(self, file):
        return file

    # This is a no-op except on Win32
    def SourceServerIndexing(self, debug_file, guid, sourceFileStream, vcs_root):
        return ""

    # subclasses override this if they want to support this
    def CopyDebug(self, file, debug_file, guid):
        pass

    def Finish(self, stop_pool=True):
        """Wait for the expected number of jobs to be submitted, and then
        wait for the pool to finish processing them. By default, will close
        and clear the pool, but for testcases that need multiple runs, pass
        stop_pool = False."""
        with Dumper.jobs_condition:
            while len(self.jobs_record) != 0:
                Dumper.jobs_condition.wait()
        if stop_pool:
            Dumper.pool.close()
            Dumper.pool.join()

    def Process(self, file_or_dir):
        """Process a file or all the (valid) files in a directory; processing is performed
        asynchronously, and Finish must be called to wait for it complete and cleanup."""
        if os.path.isdir(file_or_dir) and not self.ShouldSkipDir(file_or_dir):
            self.ProcessDir(file_or_dir)
        elif os.path.isfile(file_or_dir):
            self.ProcessFiles((file_or_dir,))

    def ProcessDir(self, dir):
        """Process all the valid files in this directory.  Valid files
        are determined by calling ShouldProcess; processing is performed
        asynchronously, and Finish must be called to wait for it complete and cleanup."""
        for root, dirs, files in os.walk(dir):
            for d in dirs[:]:
                if self.ShouldSkipDir(d):
                    dirs.remove(d)
            for f in files:
                fullpath = os.path.join(root, f)
                if self.ShouldProcess(fullpath):
                    self.ProcessFiles((fullpath,))

    def SubmitJob(self, file_key, func, args, callback):
        """Submits a job to the pool of workers; increments the number of submitted jobs."""
        self.JobStarted(file_key)
        res = Dumper.pool.apply_async(func, args=args, callback=callback)

    def ProcessFilesFinished(self, res):
        """Callback from multiprocesing when ProcessFilesWork finishes;
        run the cleanup work, if any"""
        self.JobFinished(res['files'][-1])
        # only run the cleanup function once per tuple of files
        self.files_record[res['files']] += 1
        if self.files_record[res['files']] == len(self.archs):
            del self.files_record[res['files']]
            if res['after']:
                res['after'](res['status'], res['after_arg'])

    def ProcessFiles(self, files, after=None, after_arg=None):
        """Dump symbols from these files into a symbol file, stored
        in the proper directory structure in  |symbol_path|; processing is performed
        asynchronously, and Finish must be called to wait for it complete and cleanup.
        All files after the first are fallbacks in case the first file does not process
        successfully; if it does, no other files will be touched."""
        self.output_pid(sys.stderr, "Submitting jobs for files: %s" % str(files))

        # tries to get the vcs root from the .mozconfig first - if it's not set
        # the tinderbox vcs path will be assigned further down
        vcs_root = os.environ.get("SRCSRV_ROOT")
        for arch_num, arch in enumerate(self.archs):
            self.files_record[files] = 0 # record that we submitted jobs for this tuple of files
            self.SubmitJob(files[-1], StartProcessFilesWork, args=(self, files, arch_num, arch, vcs_root, after, after_arg), callback=self.ProcessFilesFinished)

    def ProcessFilesWork(self, files, arch_num, arch, vcs_root, after, after_arg):
        self.output_pid(sys.stderr, "Worker processing files: %s" % (files,))

        # our result is a status, a cleanup function, an argument to that function, and the tuple of files we were called on
        result = { 'status' : False, 'after' : after, 'after_arg' : after_arg, 'files' : files }

        sourceFileStream = ''
        for file in files:
            # files is a tuple of files, containing fallbacks in case the first file doesn't process successfully
            try:
                proc = subprocess.Popen([self.dump_syms] + arch.split() + [file],
                                        stdout=subprocess.PIPE)
                module_line = proc.stdout.next()
                if module_line.startswith("MODULE"):
                    # MODULE os cpu guid debug_file
                    (guid, debug_file) = (module_line.split())[3:5]
                    # strip off .pdb extensions, and append .sym
                    sym_file = re.sub("\.pdb$", "", debug_file) + ".sym"
                    # we do want forward slashes here
                    rel_path = os.path.join(debug_file,
                                            guid,
                                            sym_file).replace("\\", "/")
                    full_path = os.path.normpath(os.path.join(self.symbol_path,
                                                              rel_path))
                    try:
                        os.makedirs(os.path.dirname(full_path))
                    except OSError: # already exists
                        pass
                    f = open(full_path, "w")
                    f.write(module_line)
                    # now process the rest of the output
                    for line in proc.stdout:
                        if line.startswith("FILE"):
                            # FILE index filename
                            (x, index, filename) = line.rstrip().split(None, 2)
                            if sys.platform == "sunos5":
                                for srcdir in self.srcdirs:
                                    start = filename.find(self.srcdir)
                                    if start != -1:
                                        filename = filename[start:]
                                        break
                            filename = self.FixFilenameCase(filename)
                            sourcepath = filename
                            if self.vcsinfo:
                                (filename, rootname) = GetVCSFilename(filename, self.srcdirs)
                                # sets vcs_root in case the loop through files were to end on an empty rootname
                                if vcs_root is None:
                                  if rootname:
                                     vcs_root = rootname
                            # gather up files with hg for indexing   
                            if filename.startswith("hg"):
                                (ver, checkout, source_file, revision) = filename.split(":", 3)
                                sourceFileStream += sourcepath + "*" + source_file + '*' + revision + "\r\n"
                            f.write("FILE %s %s\n" % (index, filename))
                        else:
                            # pass through all other lines unchanged
                            f.write(line)
                            # we want to return true only if at least one line is not a MODULE or FILE line
                            result['status'] = True
                    f.close()
                    proc.wait()
                    # we output relative paths so callers can get a list of what
                    # was generated
                    self.output(sys.stdout, rel_path)
                    if self.srcsrv and vcs_root:
                        # add source server indexing to the pdb file
                        self.SourceServerIndexing(file, guid, sourceFileStream, vcs_root)
                    # only copy debug the first time if we have multiple architectures
                    if self.copy_debug and arch_num == 0:
                        self.CopyDebug(file, debug_file, guid)
            except StopIteration:
                pass
            except e:
                self.output(sys.stderr, "Unexpected error: %s" % (str(e),))
                raise
            if result['status']:
                # we only need 1 file to work
                break
        return result

# Platform-specific subclasses.  For the most part, these just have
# logic to determine what files to extract symbols from.

class Dumper_Win32(Dumper):
    fixedFilenameCaseCache = {}

    def ShouldProcess(self, file):
        """This function will allow processing of pdb files that have dll
        or exe files with the same base name next to them."""
        if not Dumper.ShouldProcess(self, file):
            return False
        if file.endswith(".pdb"):
            (path,ext) = os.path.splitext(file)
            if os.path.isfile(path + ".exe") or os.path.isfile(path + ".dll"):
                return True
        return False

    def FixFilenameCase(self, file):
        """Recent versions of Visual C++ put filenames into
        PDB files as all lowercase.  If the file exists
        on the local filesystem, fix it."""

        # Use a cached version if we have one.
        if file in self.fixedFilenameCaseCache:
            return self.fixedFilenameCaseCache[file]

        result = file

        (path, filename) = os.path.split(file)
        if os.path.isdir(path):
            lc_filename = filename.lower()
            for f in os.listdir(path):
                if f.lower() == lc_filename:
                    result = os.path.join(path, f)
                    break

        # Cache the corrected version to avoid future filesystem hits.
        self.fixedFilenameCaseCache[file] = result
        return result

    def CopyDebug(self, file, debug_file, guid):
        rel_path = os.path.join(debug_file,
                                guid,
                                debug_file).replace("\\", "/")
        full_path = os.path.normpath(os.path.join(self.symbol_path,
                                                  rel_path))
        shutil.copyfile(file, full_path)
        # try compressing it
        compressed_file = os.path.splitext(full_path)[0] + ".pd_"
        # ignore makecab's output
        success = subprocess.call(["makecab.exe", "/D", "CompressionType=LZX", "/D",
                                   "CompressionMemory=21",
                                   full_path, compressed_file],
                                  stdout=open("NUL:","w"), stderr=subprocess.STDOUT)
        if success == 0 and os.path.exists(compressed_file):
            os.unlink(full_path)
            self.output(sys.stdout, os.path.splitext(rel_path)[0] + ".pd_")
        else:
            self.output(sys.stdout, rel_path)
        
    def SourceServerIndexing(self, debug_file, guid, sourceFileStream, vcs_root):
        # Creates a .pdb.stream file in the mozilla\objdir to be used for source indexing
        debug_file = os.path.abspath(debug_file)
        streamFilename = debug_file + ".stream"
        stream_output_path = os.path.abspath(streamFilename)
        # Call SourceIndex to create the .stream file
        result = SourceIndex(sourceFileStream, stream_output_path, vcs_root)
        if self.copy_debug:
            pdbstr_path = os.environ.get("PDBSTR_PATH")
            pdbstr = os.path.normpath(pdbstr_path)
            subprocess.call([pdbstr, "-w", "-p:" + os.path.basename(debug_file),
                             "-i:" + os.path.basename(streamFilename), "-s:srcsrv"],
                            cwd=os.path.dirname(stream_output_path))
            # clean up all the .stream files when done
            os.remove(stream_output_path)
        return result

class Dumper_Linux(Dumper):
    objcopy = os.environ['OBJCOPY'] if 'OBJCOPY' in os.environ else 'objcopy'
    def ShouldProcess(self, file):
        """This function will allow processing of files that are
        executable, or end with the .so extension, and additionally
        file(1) reports as being ELF files.  It expects to find the file
        command in PATH."""
        if not Dumper.ShouldProcess(self, file):
            return False
        if file.endswith(".so") or os.access(file, os.X_OK):
            return self.RunFileCommand(file).startswith("ELF")
        return False

    def CopyDebug(self, file, debug_file, guid):
        # We want to strip out the debug info, and add a
        # .gnu_debuglink section to the object, so the debugger can
        # actually load our debug info later.
        file_dbg = file + ".dbg"
        if subprocess.call([self.objcopy, '--only-keep-debug', file, file_dbg]) == 0 and \
           subprocess.call([self.objcopy, '--add-gnu-debuglink=%s' % file_dbg, file]) == 0:
            rel_path = os.path.join(debug_file,
                                    guid,
                                    debug_file + ".dbg")
            full_path = os.path.normpath(os.path.join(self.symbol_path,
                                                      rel_path))
            shutil.move(file_dbg, full_path)
            # gzip the shipped debug files
            os.system("gzip %s" % full_path)
            self.output(sys.stdout, rel_path + ".gz")
        else:
            if os.path.isfile(file_dbg):
                os.unlink(file_dbg)

class Dumper_Solaris(Dumper):
    def RunFileCommand(self, file):
        """Utility function, returns the output of file(1)"""
        try:
            output = os.popen("file " + file).read()
            return output.split('\t')[1];
        except:
            return ""

    def ShouldProcess(self, file):
        """This function will allow processing of files that are
        executable, or end with the .so extension, and additionally
        file(1) reports as being ELF files.  It expects to find the file
        command in PATH."""
        if not Dumper.ShouldProcess(self, file):
            return False
        if file.endswith(".so") or os.access(file, os.X_OK):
            return self.RunFileCommand(file).startswith("ELF")
        return False

def StartProcessFilesWorkMac(dumper, file):
    """multiprocessing can't handle methods as Process targets, so we define
    a simple wrapper function around the work method."""
    return dumper.ProcessFilesWorkMac(file)

def AfterMac(status, dsymbundle):
    """Cleanup function to run on Macs after we process the file(s)."""
    # CopyDebug will already have been run from Dumper.ProcessFiles
    shutil.rmtree(dsymbundle)

class Dumper_Mac(Dumper):
    def ShouldProcess(self, file):
        """This function will allow processing of files that are
        executable, or end with the .dylib extension, and additionally
        file(1) reports as being Mach-O files.  It expects to find the file
        command in PATH."""
        if not Dumper.ShouldProcess(self, file):
            return False
        if file.endswith(".dylib") or os.access(file, os.X_OK):
            return self.RunFileCommand(file).startswith("Mach-O")
        return False

    def ShouldSkipDir(self, dir):
        """We create .dSYM bundles on the fly, but if someone runs
        buildsymbols twice, we should skip any bundles we created
        previously, otherwise we'll recurse into them and try to 
        dump the inner bits again."""
        if dir.endswith(".dSYM"):
            return True
        return False

    def ProcessFiles(self, files, after=None, after_arg=None):
        # also note, files must be len 1 here, since we're the only ones
        # that ever add more than one file to the list
        self.output_pid(sys.stderr, "Submitting job for Mac pre-processing on file: %s" % (files[0]))
        self.SubmitJob(files[0], StartProcessFilesWorkMac, args=(self, files[0]), callback=self.ProcessFilesMacFinished)

    def ProcessFilesMacFinished(self, result):
        if result['status']:
            # kick off new jobs per-arch with our new list of files
            Dumper.ProcessFiles(self, result['files'], after=AfterMac, after_arg=result['files'][0])
        # only decrement jobs *after* that, since otherwise we'll remove the record for this file
        self.JobFinished(result['files'][-1])

    def ProcessFilesWorkMac(self, file):
        """dump_syms on Mac needs to be run on a dSYM bundle produced
        by dsymutil(1), so run dsymutil here and pass the bundle name
        down to the superclass method instead."""
        self.output_pid(sys.stderr, "Worker running Mac pre-processing on file: %s" % (file,))

        # our return is a status and a tuple of files to dump symbols for
        # the extra files are fallbacks; as soon as one is dumped successfully, we stop
        result = { 'status' : False, 'files' : None, 'file_key' : file }
        dsymbundle = file + ".dSYM"
        if os.path.exists(dsymbundle):
            shutil.rmtree(dsymbundle)
        # dsymutil takes --arch=foo instead of -a foo like everything else
        subprocess.call(["dsymutil"] + [a.replace('-a ', '--arch=') for a in self.archs if a]
                        + [file],
                        stdout=open("/dev/null","w"))
        if not os.path.exists(dsymbundle):
            # dsymutil won't produce a .dSYM for files without symbols
            result['status'] = False
            return result

        result['status'] = True
        result['files'] = (dsymbundle, file)
        return result

    def CopyDebug(self, file, debug_file, guid):
        """ProcessFiles has already produced a dSYM bundle, so we should just
        copy that to the destination directory. However, we'll package it
        into a .tar.bz2 because the debug symbols are pretty huge, and
        also because it's a bundle, so it's a directory. |file| here is the
        dSYM bundle, and |debug_file| is the original filename."""
        rel_path = os.path.join(debug_file,
                                guid,
                                os.path.basename(file) + ".tar.bz2")
        full_path = os.path.abspath(os.path.join(self.symbol_path,
                                                  rel_path))
        success = subprocess.call(["tar", "cjf", full_path, os.path.basename(file)],
                                  cwd=os.path.dirname(file),
                                  stdout=open("/dev/null","w"), stderr=subprocess.STDOUT)
        if success == 0 and os.path.exists(full_path):
            self.output(sys.stdout, rel_path)

# Entry point if called as a standalone program
def main():
    parser = OptionParser(usage="usage: %prog [options] <dump_syms binary> <symbol store path> <debug info files>")
    parser.add_option("-c", "--copy",
                      action="store_true", dest="copy_debug", default=False,
                      help="Copy debug info files into the same directory structure as symbol files")
    parser.add_option("-a", "--archs",
                      action="store", dest="archs",
                      help="Run dump_syms -a <arch> for each space separated cpu architecture in ARCHS (only on OS X)")
    parser.add_option("-s", "--srcdir",
                      action="append", dest="srcdir", default=[],
                      help="Use SRCDIR to determine relative paths to source files")
    parser.add_option("-v", "--vcs-info",
                      action="store_true", dest="vcsinfo",
                      help="Try to retrieve VCS info for each FILE listed in the output")
    parser.add_option("-i", "--source-index",
                      action="store_true", dest="srcsrv", default=False,
                      help="Add source index information to debug files, making them suitable for use in a source server.")
    parser.add_option("-x", "--exclude",
                      action="append", dest="exclude", default=[], metavar="PATTERN",
                      help="Skip processing files matching PATTERN.")
    parser.add_option("--repo-manifest",
                      action="store", dest="repo_manifest",
                      help="""Get source information from this XML manifest
produced by the `repo manifest -r` command.
""")
    (options, args) = parser.parse_args()

    #check to see if the pdbstr.exe exists
    if options.srcsrv:
        pdbstr = os.environ.get("PDBSTR_PATH")
        if not os.path.exists(pdbstr):
            print >> sys.stderr, "Invalid path to pdbstr.exe - please set/check PDBSTR_PATH.\n"
            sys.exit(1)

    if len(args) < 3:
        parser.error("not enough arguments")
        exit(1)

    dumper = GetPlatformSpecificDumper(dump_syms=args[0],
                                       symbol_path=args[1],
                                       copy_debug=options.copy_debug,
                                       archs=options.archs,
                                       srcdirs=options.srcdir,
                                       vcsinfo=options.vcsinfo,
                                       srcsrv=options.srcsrv,
                                       exclude=options.exclude,
                                       repo_manifest=options.repo_manifest)
    for arg in args[2:]:
        dumper.Process(arg)
    dumper.Finish()

# run main if run directly
if __name__ == "__main__":
    # set up the multiprocessing infrastructure before we start;
    # note that this needs to be in the __main__ guard, or else Windows will choke
    Dumper.GlobalInit()

    main()
